// ----------------------------------------------------------------------------------------------
// Copyright (c) fulesnabel@hotmail.se
// ----------------------------------------------------------------------------------------------
// This source code is subject to terms and conditions of the Microsoft Public License. A 
// copy of the license can be found in the License.html file at the root of this distribution. 
// If you cannot locate the  Microsoft Public License, please send an email to 
// dlr@microsoft.com. By using this source code in any fashion, you are agreeing to be bound 
//  by the terms of the Microsoft Public License.
// ----------------------------------------------------------------------------------------------
// You must not remove this notice, or any other, from this software.
// ----------------------------------------------------------------------------------------------

namespace SBFastGraphics.Internal
{
    using System.Collections.Generic;
    using System.Diagnostics;
    using System.Threading;

    sealed class ConcurrentQueue<T> : BaseDisposable
    {
        static ConcurrentQueue ()
        {
            s_timer.Start ();
        }

        static readonly Stopwatch s_timer = new Stopwatch ();

        const int InitialSize = 128;
        readonly object m_lock = new object ();
        volatile Queue<T> m_queue = new Queue<T> (InitialSize);
        volatile bool m_isTerminated;

        public bool IsTerminated
        {
            get
            {
                return m_isTerminated;
            }
        }

        public bool Enqueue (T value)
        {
            lock (m_lock)
            {
                if (m_isTerminated)
                {
                    return false;
                }

                m_queue.Enqueue (value);
                Monitor.Pulse (m_lock);
                return true;
            }
        }

        bool DetachQueue_NoLock (out Queue<T> queue)
        {
            if (m_queue.Count > 0)
            {
                queue = m_queue;
                m_queue = new Queue<T> (InitialSize);
                return true;
            }

            queue = null;
            return false;
        }

        public bool DetachQueue (out Queue<T> queue)
        {
            lock (m_lock)
            {
                return DetachQueue_NoLock (out queue);
            }
        }

        public bool DetachQueue (int timoutInMilliseconds, out Queue<T> queue)
        {
            lock (m_lock)
            {
                var result = DetachQueue_NoLock (out queue);
                if (result)
                {
                    return true;
                }

                var begin = s_timer.ElapsedMilliseconds;
                var now = begin;
                var end = timoutInMilliseconds + begin;

                while (
                        end > (now = s_timer.ElapsedMilliseconds)
                    && Monitor.Wait (m_lock, (int)(end - now))
                    )
                {
                    if (DetachQueue_NoLock (out queue))
                    {
                        return true;
                    }
                }

                return false;
            }
        }

        public void FlushAndTerminateQueue (params T[] terminateMessages)
        {
            lock (m_lock)
            {
                m_isTerminated = true;
                m_queue.Clear ();
                foreach (var terminateMessage in terminateMessages)
                {
                    m_queue.Enqueue (terminateMessage);
                }
            }
        }

        protected override void OnDispose ()
        {
            FlushAndTerminateQueue ();
        }
    }
}