// ----------------------------------------------------------------------------------------------
// Copyright (c) fulesnabel@hotmail.se
// ----------------------------------------------------------------------------------------------
// This source code is subject to terms and conditions of the Microsoft Public License. A 
// copy of the license can be found in the License.html file at the root of this distribution. 
// If you cannot locate the  Microsoft Public License, please send an email to 
// dlr@microsoft.com. By using this source code in any fashion, you are agreeing to be bound 
//  by the terms of the Microsoft Public License.
// ----------------------------------------------------------------------------------------------
// You must not remove this notice, or any other, from this software.
// ----------------------------------------------------------------------------------------------

namespace SBFastGraphics.Internal
{
    using System;
    using System.Collections.Generic;
    using System.Globalization;
    using System.Text.RegularExpressions;
    using Microsoft.WindowsAPICodePack.DirectX.Direct2D1;

    static partial class Extensions
    {
        static readonly CultureInfo s_culture = CultureInfo.InvariantCulture;

        // --------------------------------------------------------------------
        // string
        // --------------------------------------------------------------------
        static readonly Regex s_colorMatch = new Regex (
            @"^\#(?<a>[0-9,a-f,A-F]{2})?(?<c>[0-9,a-f,A-F]{6})$",
            RegexOptions.Compiled |  RegexOptions.CultureInvariant
            );
        public static ColorF ParseAsColor (this string s)
        {
            var match = s_colorMatch.Match (s.DefaultTo ());
            if (!match.Success)
            {
                return new ColorF (1, 0, 0);
            }


            var alpha = 1.0F;
            var alphaMatch = match.Groups["a"];
            if (alphaMatch.Success)
            {
                alpha = alphaMatch.Value.ParseAsHex (255)/255.0F;
            }

            var color = match.Groups["c"].Value.ParseAsHex (0);

            var red = (0xFF & (color >> (8*2)))/255.0F;
            var green = (0xFF & (color >> (8*1)))/255.0F;
            var blue = (0xFF & (color >> (8*0)))/255.0F;

            return new ColorF (red, green, blue, alpha);
        }

        public static TEnum ParseAsEnum<TEnum> (this string s, TEnum defaultValue = default (TEnum))
            where TEnum : struct
        {
            TEnum value;
            return EnumParser<TEnum>.TryParse (s, out value) ? value : defaultValue;
        }

        public static int ParseAsHex (this string s, int defaultValue = default (int))
        {
            int value;
            return
                int.TryParse (s.DefaultTo (""), NumberStyles.HexNumber, s_culture, out value)
                    ? value
                    : defaultValue
                ;

        }

        public static string DefaultTo (this string s, string defaultValue = "")
        {
            return !string.IsNullOrEmpty (s) ? s : defaultValue;
        }

        public static bool IsNullOrEmpty (this string s)
        {
            return string.IsNullOrEmpty (s);
        }

        public static string FormatWith (this string s, params object[] args)
        {
            return string.Format (s_culture, s ?? "", args ?? new object[0]);
        }

        // --------------------------------------------------------------------

        // --------------------------------------------------------------------
        // Dictionary<TKey, TValue>
        // --------------------------------------------------------------------
        public static TValue GetOrDefault<TKey, TValue> (
            this Dictionary<TKey, TValue> dictionary,
            TKey key,
            TValue defaultValue = default (TValue)
            )
        {
            if (dictionary == null)
            {
                return defaultValue;
            }

            TValue value;
            return dictionary.TryGetValue (key, out value)
                ?   value
                :   defaultValue
                ;
        }
        // --------------------------------------------------------------------

        // --------------------------------------------------------------------
        // TValue (ref)
        // --------------------------------------------------------------------
        public static TValue DefaultTo<TValue> (this TValue v, TValue defaultValue)
            where TValue : class
        {
            return v ?? defaultValue;
        }
        // --------------------------------------------------------------------

        // --------------------------------------------------------------------
        // object
        // --------------------------------------------------------------------
        public static void SafeDispose (this object value)
        {
            var disposable = value as IDisposable;

            if (disposable != null)
            {
                try
                {
                    disposable.Dispose ();
                }
                catch (Exception exc)
                {
                    Log.Error ("Failed to dispose {0}: {1}", disposable.GetType ().FullName, exc);
                }
            }
        }
        // --------------------------------------------------------------------
    }
}