// ----------------------------------------------------------------------------------------------
// Copyright (c) fulesnabel@hotmail.se
// ----------------------------------------------------------------------------------------------
// This source code is subject to terms and conditions of the Microsoft Public License. A 
// copy of the license can be found in the License.html file at the root of this distribution. 
// If you cannot locate the  Microsoft Public License, please send an email to 
// dlr@microsoft.com. By using this source code in any fashion, you are agreeing to be bound 
//  by the terms of the Microsoft Public License.
// ----------------------------------------------------------------------------------------------
// You must not remove this notice, or any other, from this software.
// ----------------------------------------------------------------------------------------------

namespace SBFastGraphics.Internal
{
    using System;
    using System.Collections;
    using System.Collections.Generic;
    using System.Linq;
    using Interfaces;

    sealed class LifetimeContainer<T, TContext> : BaseDisposable, IEnumerable<T>
        where T : class, ILifetime<TContext> 
        where TContext : class
    {
        readonly Dictionary<string, T> m_dic = new Dictionary<string, T> ();

        protected override void OnDispose ()
        {
            RemoveAll ();
        }

        public TValue Find<TValue> (string name) 
            where TValue : class, ILifetime<TContext>
        {
            return m_dic.GetOrDefault (name.DefaultTo ()) as TValue;
        }

        public void AddOrReplace (string name, T value, TContext context)
        {
            RemoveImpl (name);
            if (value != null)
            {
                m_dic.Add (name.DefaultTo (), value);
                value.Setup (context);
            }
        }

        void RemoveImpl (string name)
        {
            var value = m_dic.GetOrDefault (name.DefaultTo ());
            if (value != null)
            {
                RemoveImpl (value);
                m_dic.Remove (name.DefaultTo ());
            }
        }

        public void Remove (string name)
        {
            RemoveImpl (name);
        }

        static void RemoveImpl (T value)
        {
            value.SafeDispose ();
        }

        public void RemoveAll ()
        {
            try
            {
                foreach (var pair in m_dic)
                {
                    RemoveImpl (pair.Value);
                }
            }
            finally
            {
                m_dic.Clear ();
            }
        }


        public void TearDown ()
        {
            foreach (var pair in m_dic)
            {
                var renderable = pair.Value;
                try
                {
                    renderable.TearDown ();
                }
                catch (Exception exc)
                {
                    Log.Error ("TearDown failed for {0} : {1}", renderable, exc);
                }
            }
        }

        public IEnumerator<T> GetEnumerator ()
        {
            return m_dic.Select (kv => kv.Value).GetEnumerator ();
        }

        IEnumerator IEnumerable.GetEnumerator ()
        {
            return GetEnumerator ();
        }
    }
}