// ----------------------------------------------------------------------------------------------
// Copyright (c) fulesnabel@hotmail.se
// ----------------------------------------------------------------------------------------------
// This source code is subject to terms and conditions of the Microsoft Public License. A 
// copy of the license can be found in the License.html file at the root of this distribution. 
// If you cannot locate the  Microsoft Public License, please send an email to 
// dlr@microsoft.com. By using this source code in any fashion, you are agreeing to be bound 
//  by the terms of the Microsoft Public License.
// ----------------------------------------------------------------------------------------------
// You must not remove this notice, or any other, from this software.
// ----------------------------------------------------------------------------------------------

namespace SBFastGraphics.Internal
{
    using System;
    using Microsoft.WindowsAPICodePack.DirectX.Direct2D1;
    using Microsoft.WindowsAPICodePack.DirectX.DirectWrite;
    using Interfaces;

    abstract class BaseRenderable : IRenderable
    {
        RenderContext m_ctx;
        public RenderContext Context
        {
            get { return m_ctx; }
        }

        public abstract void Render ();

        public virtual void Setup (RenderContext ctx)
        {
            m_ctx = ctx;
        }

        public virtual void TearDown ()
        {
            m_ctx = null;
        }

        protected virtual void OnDispose ()
        {
            
        }

        public void Dispose ()
        {
            TearDown ();
            OnDispose ();
        }
    }

    class RenderContext
    {
        public readonly D2DFactory Direct2D;
        public readonly DWriteFactory DirectWrite;

        public readonly RenderTarget Target;
        readonly Brush m_brush;
        readonly LifetimeContainer<IResource, ResourceContext> m_resources;

        public RenderContext (
            D2DFactory direct2D, 
            DWriteFactory directWrite,
            RenderTarget target, 
            Brush brush,
            LifetimeContainer<IResource, ResourceContext> resources
            )
        {
            Direct2D = direct2D;
            DirectWrite = directWrite;
            Target = target;
            DirectWrite = directWrite;
            Direct2D = direct2D;
            m_brush = brush;
            m_resources = resources;
        }

        public Brush GetBrushAndAdjustToRectangularCoordinates (string brushName, float opacity, float x, float y, float width, float height)
        {
            var brush = m_resources.Find<IBrush> (brushName.DefaultTo ());
            var b = brush != null ? brush.Adjust (x, y, x + width, y + height) : m_brush;
            b.Opacity = opacity;
            return b;
        }

        public Brush GetBrushAndAdjustToPolarCoordinates (string brushName, float opacity, float centerX, float centerY, float radiusX, float radiusY)
        {
            var brush = m_resources.Find<IBrush> (brushName.DefaultTo ());

            var x0 = centerX - radiusX;
            var y0 = centerY - radiusY;
            var x1 = centerX + radiusX;
            var y1 = centerY + radiusY;

            var b =  brush != null ? brush.Adjust (x0, y0, x1, y1) ?? m_brush : m_brush;
            b.Opacity = opacity;
            return b;
        }

        public D2DBitmap GetBitmap (string bitmapName)
        {
            var bmp = m_resources.Find<IBitmap> (bitmapName.DefaultTo ());
            return bmp != null ? bmp.Bitmap : null;
        }
    }

    partial class Line : BaseRenderable
    {
        public override void Render ()
        {
            Context.Target.DrawLine (
                new Point2F (X0, Y0),
                new Point2F (X1, Y1),
                Context.GetBrushAndAdjustToRectangularCoordinates (Brush, Opacity, X0, Y0, X1 - X0, Y1 - Y0),
                Thickness
                );
        }
    }

    partial class Rectangle : BaseRenderable
    {
        public override void Render ()
        {
            Context.Target.DrawRectangle (
                new RectF (X, Y, X + Width, Y + Height),
                Context.GetBrushAndAdjustToRectangularCoordinates (Brush, Opacity, X, Y, Width, Height),
                Thickness
                );
        }
    }

    partial class FilledRectangle : BaseRenderable
    {
        public override void Render ()
        {
            Context.Target.FillRectangle (
                new RectF (X, Y, X + Width, Y + Height),
                Context.GetBrushAndAdjustToRectangularCoordinates (Brush, Opacity, X, Y, Width, Height)
                );
        }
    }

    abstract class BaseGeometry : BaseRenderable
    {
        public Geometry Geometry;

        public override void Setup (RenderContext ctx)
        {
            base.Setup (ctx);
            Geometry = CreateGeometry (ctx);
        }

        public override void TearDown ()
        {
            base.TearDown ();
            Geometry.SafeDispose ();
            Geometry = null;
        }

        protected abstract Geometry CreateGeometry (RenderContext ctx);
    }

    partial class RoundedRectangle : BaseRenderable
    {
        public override void Render ()
        {
            Context.Target.DrawRoundedRectangle (
                new  RoundedRect (new RectF (X, Y, X + Width, Y + Height), RadiusX, RadiusY),
                Context.GetBrushAndAdjustToRectangularCoordinates (Brush, Opacity, X, Y, Width, Height),
                Thickness
                );
        }
    }

    partial class FilledRoundedRectangle : BaseRenderable
    {
        public override void Render ()
        {
            Context.Target.FillRoundedRectangle (
                new RoundedRect (new RectF (X, Y, X + Width, Y + Height), RadiusX, RadiusY),
                Context.GetBrushAndAdjustToRectangularCoordinates (Brush, Opacity, X, Y, Width, Height)
                );
        }
    }

    partial class Ellipse : BaseRenderable
    {
        public override void Render ()
        {
            Context.Target.DrawEllipse (
                new Microsoft.WindowsAPICodePack.DirectX.Direct2D1.Ellipse (
                    new Point2F (CenterX, CenterY),
                    RadiusX,
                    RadiusY),
                Context.GetBrushAndAdjustToPolarCoordinates (Brush, Opacity, CenterX, CenterY, RadiusX, RadiusY),
                Thickness
                );
        }
    }

    partial class FilledEllipse : BaseRenderable
    {
        public override void Render ()
        {
            Context.Target.FillEllipse (
                new Microsoft.WindowsAPICodePack.DirectX.Direct2D1.Ellipse (
                    new Point2F (CenterX, CenterY),
                    RadiusX,
                    RadiusY),
                Context.GetBrushAndAdjustToPolarCoordinates (Brush, Opacity, CenterX, CenterY, RadiusX, RadiusY)
                );
        }
    }

    partial class TextBlock : BaseRenderable
    {
        public TextLayout TextLayout;

        public override void Render ()
        {
            if (!Message.IsNullOrEmpty () && TextLayout != null)
            {
                Context.Target.DrawTextLayout (
                    new Point2F (X, Y),
                    TextLayout,
                    Context.GetBrushAndAdjustToRectangularCoordinates (Brush, Opacity, X, Y, Width, Height),
                    DrawTextOptions.Clip
                    );
            }
        }

        public override void Setup (RenderContext ctx)
        {
            base.Setup (ctx);

            if (TextLayout == null)
            {
                TextLayout = ctx.DirectWrite.CreateTextLayout (
                    Message.DefaultTo (),
                    ctx.DirectWrite.CreateTextFormat (
                        FontFamily.DefaultTo ("Courier New"),
                        Math.Max (4.0F, FontSize)
                        ),
                    Width,
                    Height
                    );
            }

        }

        protected override void OnDispose ()
        {
            TextLayout.SafeDispose ();
            TextLayout = null;
        }
    }


    partial class Bitmap : BaseRenderable
    {
        public override void Render ()
        {
            var bitmap = Context.GetBitmap (BitmapName);
            if (bitmap != null)
            {
                Context.Target.DrawBitmap (
                    bitmap,
                    Opacity,
                    BitmapInterpolationMode.Linear,
                    new RectF (X, Y, X + Width, Y + Height)
                    );
            }
        }
    }
}