﻿// ----------------------------------------------------------------------------------------------
// Copyright (c) fulesnabel@hotmail.se
// ----------------------------------------------------------------------------------------------
// This source code is subject to terms and conditions of the Microsoft Public License. A 
// copy of the license can be found in the License.html file at the root of this distribution. 
// If you cannot locate the  Microsoft Public License, please send an email to 
// dlr@microsoft.com. By using this source code in any fashion, you are agreeing to be bound 
//  by the terms of the Microsoft Public License.
// ----------------------------------------------------------------------------------------------
// You must not remove this notice, or any other, from this software.
// ----------------------------------------------------------------------------------------------

namespace SBFastGraphics.Internal
{
    using System;
    using System.Collections.Generic;
    using System.IO;
    using System.Linq;
    using System.Net;
    using Microsoft.WindowsAPICodePack.DirectX.Direct2D1;
    using Microsoft.WindowsAPICodePack.DirectX.DirectWrite;
    using Microsoft.WindowsAPICodePack.DirectX.WindowsImagingComponent;
    using Interfaces;

    class ResourceContext
    {
        public readonly D2DFactory Direct2D;
        public readonly DWriteFactory DirectWrite;
        public readonly ImagingFactory ImagingFactory;

        public readonly RenderTarget Target;

        public ResourceContext (
            D2DFactory direct2D,
            DWriteFactory directWrite,
            ImagingFactory imagingFactory,
            RenderTarget target
            )
        {
            Direct2D = direct2D;
            DirectWrite = directWrite;
            ImagingFactory = imagingFactory;
            Target = target;
        }
    }

    abstract class BaseResource : IResource
    {
        ResourceContext m_ctx;
        public ResourceContext Context
        {
            get
            {
                return m_ctx;
            }
        }

        public virtual void Setup (ResourceContext ctx)
        {
            m_ctx = ctx;
        }

        public virtual void TearDown ()
        {
            m_ctx = null;
        }

        protected virtual void OnDispose ()
        {

        }

        public void Dispose ()
        {
            TearDown ();
            OnDispose ();
        }
    }

    abstract class BaseBrush : BaseResource, IBrush
    {
        Brush m_brush;
        public Brush Brush
        {
            get
            {
                if (m_brush == null)
                {
                    m_brush = CreateBrush ();
                }
                return m_brush;
            }
            set
            {
                m_brush = value;
            }
        }


        protected abstract Brush CreateBrush ();

        public override void Setup (ResourceContext ctx)
        {
            base.Setup (ctx);
            InvalidateBrush ();
        }

        public override void TearDown ()
        {
            InvalidateBrush ();
            base.TearDown ();
        }

        public abstract Brush Adjust (float x0, float y0, float x1, float y1);

        protected void InvalidateBrush ()
        {
            if (m_brush != null)
            {
                Log.Info ("Invalidated brush for: {0}", this);
                m_brush.SafeDispose ();
                m_brush = null;
            }
        }
    }

    abstract class BaseGradientBrush : BaseBrush, IGradientBrush
    {
        public readonly List<GradientStop> Stops = new List<GradientStop> ();

        public void AddGradientStop (GradientStop stop)
        {
            Log.Info ("Added gradient stop to {0}: {1}", this, stop);

            InvalidateBrush ();
            Stops.Add (stop);
        }

        protected Microsoft.WindowsAPICodePack.DirectX.Direct2D1.GradientStop[] CreateGradientStops ()
        {
            return Stops
                .Select (
                    stop =>
                    new Microsoft.WindowsAPICodePack.DirectX.Direct2D1.GradientStop (
                        stop.Position,
                        stop.Color.ParseAsColor ()))
                .ToArray ();
        }

        protected GradientStopCollection CreateGradientStopCollection (string mode)
        {
            var stops = CreateGradientStops ();
            return Context
                .Target
                .CreateGradientStopCollection (
                    stops,
                    Gamma.StandardRgb,
                    mode.ParseAsEnum (ExtendMode.Clamp)
                );
        }
    }

    partial class GradientBrush : BaseGradientBrush
    {
        public override Brush Adjust (float x0, float y0, float x1, float y1)
        {
            var brush = Brush as LinearGradientBrush;
            if (brush != null)
            {
                var width = x1 - x0;
                var height = y1 - y0;
                if (CoordinateSpace.Equals ("Absolute", StringComparison.OrdinalIgnoreCase))
                {
                    brush.StartPoint = new Point2F (X0, Y0);
                    brush.EndPoint = new Point2F (X1, Y1);
                }
                else
                {
                    brush.StartPoint = new Point2F (x0 + X0 * width, y0 + Y0 * height);
                    brush.EndPoint = new Point2F (x0 + X1 * width, y0 + Y1 * height);
                }

            }
            return brush;
        }

        protected override Brush CreateBrush ()
        {
            if (Stops.Count == 0)
            {
                return null;
            }

            using (var gradientStopCollection = CreateGradientStopCollection (ExtendMode))
            {
                return Context.Target.CreateLinearGradientBrush (
                    new LinearGradientBrushProperties (
                        new Point2F (X0, Y0),
                        new Point2F (X1, Y1)),
                    gradientStopCollection
                    );
            }
        }
    }

    partial class CircularGradientBrush : BaseGradientBrush
    {
        public override Brush Adjust (float x0, float y0, float x1, float y1)
        {
            if (Stops.Count == 0)
            {
                return null;
            }

            var brush = Brush as RadialGradientBrush;
            if (brush != null)
            {
                var width = x1 - x0;
                var height = y1 - y0;
                var center = new Point2F ((x0 + x1) / 2, (y0 + y1) / 2);
                brush.Center = center;
                brush.RadiusX = width / 2;
                brush.RadiusY = height / 2;
                brush.GradientOriginOffset = new Point2F (OffsetX * width, OffsetY * height);
            }
            return brush;

        }

        protected override Brush CreateBrush ()
        {
            using (var gradientStopCollection = CreateGradientStopCollection (ExtendMode))
            {
                return Context.Target.CreateRadialGradientBrush (
                    new RadialGradientBrushProperties (
                        new Point2F (0.5F, 0.5F),
                        new Point2F (OffsetX, OffsetY),
                        1,
                        1),
                    gradientStopCollection
                    );
            }
        }
    }

    partial class SolidBrush : BaseBrush
    {
        protected override Brush CreateBrush ()
        {
            return Context.Target.CreateSolidColorBrush (Color.ParseAsColor ());
        }

        public override Brush Adjust (float x0, float y0, float x1, float y1)
        {
            return Brush;
        }
    }

    enum BitmapState
    {
        Uninitialized     = 0,
        Loading     = 1,
        Loaded      = 2,
        Failed      = 3,
    }

    abstract class BaseBitmap : BaseResource, IBitmap
    {
        BitmapState m_state;

        Stream m_stream;
        BitmapDecoder m_decoder;
        BitmapSource m_bitmapSource;

        D2DBitmap m_bitmap;

        public override void Setup (ResourceContext ctx)
        {
            base.Setup (ctx);

            if (m_state == BitmapState.Uninitialized)
            {
                try
                {
                    m_state = BitmapState.Loading;

                    m_stream = GetBitmapStream ();
                    if (m_stream == null)
                    {
                        m_state = BitmapState.Failed;
                        Log.Error ("Failed to get bitmap stream for: {0}", this);
                    }

                    m_decoder = ctx.ImagingFactory.CreateDecoderFromStream (m_stream, DecodeMetadataCacheOption.OnLoad);
                    {
                        if (m_decoder.FrameCount > 0)
                        {
                            m_bitmapSource = m_decoder.GetFrame (0).ToBitmapSource ();
                            m_state = BitmapState.Loaded;
                        }
                    }
                }
                catch (Exception exc)
                {
                    m_state = BitmapState.Failed;
                    Log.Error ("LoadBitmapFromFile.Setup : Caught exception: {0}", exc);
                }
            }

            if (m_bitmapSource != null)
            {
                using (var formatConverter = Context.ImagingFactory.CreateFormatConverter ())
                {
                    formatConverter.Initialize (m_bitmapSource, PixelFormats.Pbgra32Bpp, BitmapDitherType.None, BitmapPaletteType.Custom);
                    using (var convertedBitmapSource = formatConverter.ToBitmapSource ())
                    {
                        m_bitmap = ctx.Target.CreateBitmapFromWicBitmap (convertedBitmapSource);
                    }

                }
            }
        }

        protected abstract Stream GetBitmapStream ();

        public override void TearDown ()
        {
            base.TearDown ();
            m_bitmap.SafeDispose ();
            m_bitmap = null;
        }

        protected override void OnDispose ()
        {
            base.OnDispose ();
            m_stream.SafeDispose ();
            m_decoder.SafeDispose ();
            m_bitmapSource.SafeDispose ();
            m_bitmap.SafeDispose ();

            m_stream = null;
            m_decoder = null;
            m_bitmapSource = null;
            m_bitmap = null;
        }

        public D2DBitmap Bitmap
        {
            get { return m_bitmap; }
        }
    }

    partial class LoadBitmapFromFile : BaseBitmap
    {
        protected override Stream GetBitmapStream ()
        {
            var fileName = FileName ?? "";
            return File.Exists (fileName) 
                ? File.OpenRead (fileName) 
                : null
                ;
        }
    }

    partial class LoadBitmapFromUrl : BaseBitmap
    {
        protected override Stream GetBitmapStream ()
        {
            var url = URL;
            var wc  = new WebClient ();
            var bytes = wc.DownloadData (url);
            return new MemoryStream (bytes);
        }
    }
}
