// ----------------------------------------------------------------------------------------------
// Copyright (c) fulesnabel@hotmail.se
// ----------------------------------------------------------------------------------------------
// This source code is subject to terms and conditions of the Microsoft Public License. A 
// copy of the license can be found in the License.html file at the root of this distribution. 
// If you cannot locate the  Microsoft Public License, please send an email to 
// dlr@microsoft.com. By using this source code in any fashion, you are agreeing to be bound 
//  by the terms of the Microsoft Public License.
// ----------------------------------------------------------------------------------------------
// You must not remove this notice, or any other, from this software.
// ----------------------------------------------------------------------------------------------

// ReSharper disable InconsistentNaming

namespace SBFastGraphics.Internal
{
    using System;
    using System.Collections.Generic;
    using System.Diagnostics;
    using System.Drawing;
    using System.Windows.Forms;
    using Microsoft.WindowsAPICodePack.DirectX.Direct2D1;
    using Microsoft.WindowsAPICodePack.DirectX.Direct3D;
    using Microsoft.WindowsAPICodePack.DirectX.DirectWrite;
    using Microsoft.WindowsAPICodePack.DirectX.Graphics;
    using Microsoft.WindowsAPICodePack.DirectX.WindowsImagingComponent;
    using Interfaces;

    sealed partial class VisualizeForm : Form
    {
        sealed class Storage : BaseDisposable
        {
            public ColorF Background;
            public readonly LifetimeContainer<IRenderable, RenderContext> Renderable = new LifetimeContainer<IRenderable, RenderContext> ();
            public readonly LifetimeContainer<IResource, ResourceContext> Resource = new LifetimeContainer<IResource, ResourceContext> ();

            protected override void OnDispose ()
            {
                Renderable.SafeDispose ();
                Resource.SafeDispose ();
            }
        }

        Storage m_storage = new Storage ();

        readonly Timer m_startUpTimer;

        D2DFactory m_d2D;
        DWriteFactory m_dw;
        ImagingFactory m_if;
        DeviceDependentResources m_ddr;

        sealed class DeviceDependentResources : BaseDisposable
        {
            public readonly HwndRenderTarget RenderTarget;
            public readonly SolidColorBrush DefaultBrush;

            public readonly RenderContext RenderContext;
            public readonly ResourceContext ResourceContext;

            public DeviceDependentResources (
                D2DFactory d2DFactory,
                DWriteFactory dWriteFactory,
                ImagingFactory imagingFactory,
                HwndRenderTarget renderTarget,
                LifetimeContainer<IResource, ResourceContext> resources
                )
            {
                RenderTarget = renderTarget;
                DefaultBrush = renderTarget.CreateSolidColorBrush (new ColorF (1, 1, 1, 1));
                RenderContext = new RenderContext (
                    d2DFactory, 
                    dWriteFactory, 
                    RenderTarget, 
                    DefaultBrush, 
                    resources
                    );
                ResourceContext = new ResourceContext (
                    d2DFactory, 
                    dWriteFactory, 
                    imagingFactory,
                    renderTarget
                    );
            }

            protected override void OnDispose ()
            {
                RenderTarget.SafeDispose ();
                DefaultBrush.SafeDispose ();
            }
        }


        bool m_isDirty;

        readonly Stopwatch m_stopWatch = new Stopwatch ();
        readonly ConcurrentQueue<IMessage> m_queue;
        bool m_terminate;
        int m_frameCount;

        public VisualizeForm (ConcurrentQueue<IMessage> queue)
        {
            Log.Info ("VisualizeForm created");
            m_queue = queue ?? new ConcurrentQueue<IMessage> ();

            MinimumSize = new Size (100, 100);
            const int desiredFPS = 120;

            m_d2D = D2DFactory.CreateFactory (D2DFactoryType.SingleThreaded);
            m_dw = DWriteFactory.CreateFactory (DWriteFactoryType.Isolated);
            m_if = ImagingFactory.Create ();

            m_startUpTimer = new Timer { Interval = 1000 / desiredFPS };
            m_startUpTimer.Tick += StartUpTick;
            m_startUpTimer.Enabled = true;

        }

        void StartUpTick (object sender, EventArgs e)
        {
            m_startUpTimer.Enabled = false;

            while (!m_terminate)
            {
                Queue<IMessage> messages;
                if (m_queue.DetachQueue (out messages) && messages != null)
                {
                    foreach (var message in messages)
                    {
                        if (message != null)
                        {
                            try
                            {
                                ProcessMessage (message);
                            }
                            catch (Exception exc)
                            {
                                Log.Error ("Failed to process message {0}: {1}", message, exc);
                            }
                        }
                    }
                }

                Application.DoEvents ();

                if (!m_terminate)
                {
                    DrawScene ();
                }
            }

            DialogResult = DialogResult.Abort;
            Close ();
        }

        protected override void OnPaint (PaintEventArgs e)
        {
            base.OnPaint (e);

            DrawScene ();
        }

        public void DrawSceneIfDirty ()
        {
            if (m_isDirty)
            {
                DrawScene ();
            }
        }

        void IsDirty ()
        {
            m_isDirty = true;
        }

        void DrawScene ()
        {
            const int WaitFrameCount = 60;
            ++m_frameCount;
            if (m_frameCount >= WaitFrameCount && !m_stopWatch.IsRunning)
            {
                m_stopWatch.Start ();
            }

            m_isDirty = false;

            var ddr = m_ddr;
            if (ddr != null)
            {
                var renderTarget = ddr.RenderTarget;
                var defaultBrush = ddr.DefaultBrush;

                renderTarget.BeginDraw ();
                try
                {
                    renderTarget.Clear (m_storage.Background);

                    var totalSecs = m_stopWatch.Elapsed.TotalSeconds;

                    foreach (var renderable in m_storage.Renderable)
                    {
                        renderable.Render ();
                    }

                    if (m_stopWatch.IsRunning)
                    {
                        using (var layout = m_dw.CreateTextLayout (
                            "{0:0.00}s, #{1:0}, {2:0}FPS".FormatWith (totalSecs, (m_frameCount - WaitFrameCount),
                                                                      (m_frameCount - WaitFrameCount)/totalSecs),
                            m_dw.CreateTextFormat (
                                "Courier New",
                                16
                                ),
                            Width,
                            Height
                            ))
                        {
                            renderTarget.DrawTextLayout (
                                new Point2F (),
                                layout,
                                defaultBrush
                                );
                        }
                    }
                }
                finally
                {
                    renderTarget.EndDraw ();
                }

            }
        }

        protected override void OnLoad (EventArgs e)
        {
            Log.Info ("VisualizeForm loaded");
            base.OnLoad (e);
        }

        protected override void OnResize (EventArgs e)
        {
            Log.Info ("VisualizeForm resized");

            base.OnResize (e);

            var ddr = m_ddr;
            if (ddr != null)
            {
                ddr.RenderTarget.Resize (GetPixelSize ());
            }
        }

        protected override void OnHandleCreated (EventArgs e)
        {
            Log.Info ("VisualizeForm handle created");
            base.OnHandleCreated (e);
            CreateDeviceDependentResources ();
        }

        protected override void OnHandleDestroyed (EventArgs e)
        {
            Log.Info ("VisualizeForm handle destroyed");
            base.OnHandleDestroyed (e);
            FreeDeviceDependentResources ();
        }

        void CreateDeviceDependentResources ()
        {
            Log.Info ("Creating device dependent resources");
            var renderTargetProperties = new RenderTargetProperties (
                RenderTargetType.Default,
                new PixelFormat (Format.Unknown, AlphaMode.Ignore),
                96,
                96,
                RenderTargetUsages.None,
                FeatureLevel.Default
                );
            var hwndRenderTargetProperties = new HwndRenderTargetProperties (
                Handle,
                GetPixelSize (),
                Microsoft.WindowsAPICodePack.DirectX.Direct2D1.PresentOptions.None
                );           

            var renderTarget = m_d2D.CreateHwndRenderTarget (
                renderTargetProperties,
                hwndRenderTargetProperties
                );

            renderTarget.AntiAliasMode = AntiAliasMode.Aliased;
            renderTarget.TextAntiAliasMode = TextAntiAliasMode.Aliased;

            m_ddr = new DeviceDependentResources (
                m_d2D,
                m_dw,
                m_if,
                renderTarget,
                m_storage.Resource
                );

            foreach (var resource in m_storage.Resource)
            {
                resource.Setup (m_ddr.ResourceContext);
            }

            foreach (var renderable in m_storage.Renderable)
            {
                renderable.Setup (m_ddr.RenderContext);
            }


        }

        SizeU GetPixelSize ()
        {
            return new SizeU ((uint)Width, (uint)Height);
        }

        void FreeDeviceDependentResources ()
        {
            Log.Info ("Freeing device dependent resources");

            m_storage.Resource.TearDown ();
            m_storage.Renderable.TearDown ();

            m_ddr.SafeDispose ();
            m_ddr = null;
        }

        void AddRenderable (string name, IRenderable renderable)
        {
            m_storage.Renderable.AddOrReplace (name, renderable, m_ddr.RenderContext);
        }

        void AddResource (string name, IResource resource)
        {
            m_storage.Resource.AddOrReplace (name, resource, m_ddr.ResourceContext);
        }

        protected override void Dispose (bool disposing)
        {
            Log.Info ("VisualizeForm disposed");
            base.Dispose (disposing);
            if (disposing)
            {
                FreeDeviceDependentResources ();
                m_storage.SafeDispose ();
                m_if.SafeDispose ();
                m_dw.SafeDispose ();
                m_d2D.SafeDispose ();
                m_storage = null;
                m_if = null;
                m_dw = null;
                m_d2D = null;
            }
        }

        partial void ProcessMessage_RemoveResource (ref RemoveResource msg)
        {
            Log.Info ("Resource removed: {0}", msg);

            m_storage.Resource.Remove (msg.Name);
        }

        partial void ProcessMessage_Remove (ref Remove msg)
        {
            m_storage.Renderable.Remove (msg.Name);
        }

        partial void ProcessMessage_WindowSize (ref WindowSize msg)
        {
            Size = new Size (
                Math.Max (MinimumSize.Width, msg.Width),
                Math.Max (MinimumSize.Height, msg.Height)
                );
        }

        partial void ProcessMessage_MoveTextBlock (ref MoveTextBlock msg)
        {
            var textBlock = m_storage.Renderable.Find<TextBlock> (msg.Name);
            if (textBlock != null)
            {
                textBlock.X = msg.X;
                textBlock.Y = msg.Y;
            }

        }

        partial void ProcessMessage_GradientStop (ref GradientStop msg)
        {
            var gradientBrush = m_storage.Resource.Find<IGradientBrush> (msg.Name);
            if (gradientBrush != null)
            {
                gradientBrush.AddGradientStop (msg);
            }
            
        }

        partial void ProcessMessage_WindowBackground (ref WindowBackground msg)
        {
            m_storage.Background = msg.Color.ParseAsColor ();
        }

        partial void ProcessMessage_Terminate (ref Terminate msg)
        {
            m_terminate = true;
        }

        partial void ProcessMessage_RemoveAll (ref RemoveAll msg)
        {
            m_storage.Renderable.RemoveAll ();
        }

        partial void ProcessMessage_RemoveAllResources (ref RemoveAllResources msg)
        {
            m_storage.Resource.RemoveAll ();
        }

    }
}